import os, sys
sys.dont_write_bytecode = True
import numpy as np
import pandas as pd
from scipy import stats
import matplotlib.pyplot as plt
from tqdm import tqdm


py_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(py_dir)


#name = 'NXF1'
#name = 'CD40'
name = 'PIK3R3'
#name = 'IRF7'
#name = 'HLA-DRB1'
#name = 'HLA-C'

mode = 'quantity'
#mode = 'profile'

pct_threshold = 0.1 # counts threshold (e.g., 0.1 considers only counts above 10% of the total counts in the cluster)

dir_name = '/examples/outputs_local_clipnet_%s' % name
sequences = pd.read_csv(os.path.join(parent_dir + dir_name + '/heterozygous_pt01/%s_nfolds9' % mode, 'mave.csv'))['Sequence']

#clusters_dir = '/heterozygous_pt01/%s_nfolds9/clusters_umap_kmeans200' % mode
#clusters_dir = '/heterozygous_pt01/%s_nfolds9/clusters_umap_dbscan' % mode
clusters_dir = '/heterozygous_pt01/%s_nfolds9/clusters_hierarchical_cut2' % mode

save_dir = parent_dir + dir_name + clusters_dir
snp_df = pd.read_csv(os.path.join(parent_dir + dir_name, 'dbSNP_coords_%s.csv' % name))
mismatch_fname = 'mismatches_100pct'
mismatch_df = pd.read_csv(os.path.join(parent_dir + dir_name + clusters_dir, '%s.csv' % mismatch_fname))

if 1: # threshold mismatches based on a minimum cluster occupancy
    orig_length = len(mismatch_df)
    mismatch_df = mismatch_df.drop(mismatch_df[mismatch_df['Sum'] < 50].index)
    print('Thresholded %s clusters. Remaining clusters: %s' % (orig_length, len(mismatch_df)))

def snp_logic(allele_code, reference):
    disallowed = []
    allowed = []
    #allowed.append(reference)
    if allele_code == 'R':
        if reference == 'A':
            allowed.extend('RG')
        if reference == 'G':
            allowed.extend('RA')
        disallowed.extend('CTYKMSW')
    if allele_code == 'Y':
        if reference == 'C':
            allowed.extend('YT')
        if reference == 'T':
            allowed.extend('YC')
        disallowed.extend('AGRKMSW')
    if allele_code == 'K':
        if reference == 'G':
            allowed.extend('KT')
        if reference == 'T':
            allowed.extend('KG')
        disallowed.extend('ACRYMSW')
    if allele_code == 'M':
        if reference == 'A':
            allowed.extend('MC')
        if reference == 'C':
            allowed.extend('MA')
        disallowed.extend('GTRYKSW')
    if allele_code == 'S':
        if reference == 'C':
            allowed.extend('SG')
        if reference == 'G':
            allowed.extend('SC')
        disallowed.extend('ATRYKMW')
    if allele_code == 'W':
        if reference == 'A':
            allowed.extend('WT')
        if reference == 'T':
            allowed.extend('WA')
        disallowed.extend('CGRYKMS')
    if allele_code == 'B': # C/G/T (not A)
        disallowed.append('A')
        if reference == 'A':
            allowed.extend('CGT')
        if reference == 'C':
            allowed.extend('GT')
        if reference == 'G':
            allowed.extend('CT')
        if reference == 'T':
            allowed.extend('CG')
        allowed.extend('YKS')
        disallowed.extend('RMW')
    if allele_code == 'D': #A/G/T (not C)
        disallowed.append('C')
        if reference == 'A':
            allowed.extend('GT')
        if reference == 'C':
            allowed.extend('AGT')
        if reference == 'G':
            allowed.extend('AT')
        if reference == 'T':
            allowed.extend('AG')
        allowed.extend('RKW')
        disallowed.extend('YMS')
    if allele_code == 'H': #A/C/T (not G)
        disallowed.append('G')
        if reference == 'A':
            allowed.extend('CT')
        if reference == 'C':
            allowed.extend('AT')
        if reference == 'G':
            allowed.extend('ACT')
        if reference == 'T':
            allowed.extend('AC')
        allowed.extend('YMW')
        disallowed.extend('RKS')
    if allele_code == 'V': #A/C/G (not T)
        disallowed.append('T')
        if reference == 'A':
            allowed.extend('CG')
        if reference == 'C':
            allowed.extend('AG')
        if reference == 'G':
            allowed.extend('AC')
        if reference == 'T':
            allowed.extend('ACG')
        allowed.extend('RMS')
        disallowed.extend('YKW')
    if allele_code == 'N': #A/C/G/T
        allowed.extend('ACGTRYKMSW')
        disallowed = []

    return (allowed, disallowed)

print('test:', snp_logic('R','A'))


def fisher_exact_test(mismatch_df, snp_df):
    # perform Fisher exact test over positions (not considering nucleotide content)
    unique_values1, counts1 = np.unique(mismatch_df['Position'], return_counts=True)
    unique_values2, counts2 = np.unique(snp_df['seam_pos'], return_counts=True)
    possible_values = np.arange(0,500)
    union_values = np.union1d(unique_values1, unique_values2)
    contingency_table = np.zeros((2, 2), dtype=int)
    contingency_table[0,0] = len(np.intersect1d(unique_values1, unique_values2))
    contingency_table[1,1] = len(np.setxor1d(union_values, possible_values))
    contingency_table[1,0] = len(np.setdiff1d(unique_values1, unique_values2))
    contingency_table[0,1] = len(np.setdiff1d(unique_values2, unique_values1))
    odds_ratio, p_value = stats.fisher_exact(contingency_table)
    print('')
    print('Contingency table:')
    print(contingency_table)
    print("Odds Ratio:", odds_ratio)
    print("P-value:", p_value)
    print('')


fisher_exact_test(mismatch_df, snp_df)


def seam_snp_intersection(mismatch_df, snp_df, save_dir=None):
    mismatch_df['SNP_allele'] = ""
    mismatch_df['SNP_pos'] = False
    mismatch_df['SNP_base'] = False
    mismatch_df['SNP_conflict'] = False
    pos_matches_unique = 0
    for i in range(len(snp_df)):
        snp_position = snp_df['seam_pos'].iloc[i]
        snp_allele = snp_df['allele'].iloc[i]
        seam_positions = mismatch_df.loc[mismatch_df['Position'] == snp_position]
        if len(seam_positions) != 0:
            pos_matches_unique += 1
        for row in seam_positions.index:
            mismatch_df.loc[row,'SNP_allele'] = snp_allele
            mismatch_df.loc[row,'SNP_pos'] = True
            allowed, disallowed = snp_logic(snp_allele, seam_positions.loc[row]['Reference'])
            for char in allowed:
                seam_count = seam_positions.loc[row][char]
                if seam_count > (pct_threshold * seam_positions.loc[row]['Sum']):
                    mismatch_df.loc[row,'SNP_base'] = True
            for char in disallowed:
                seam_count = seam_positions.loc[row][char]
                if seam_count > (pct_threshold * seam_positions.loc[row]['Sum']):
                    if mismatch_df.loc[row,'SNP_base'] == True:
                        mismatch_df.loc[row, 'SNP_conflict'] = True
                    mismatch_df.loc[row,'SNP_base'] = False


    pos_matches = mismatch_df.loc[mismatch_df['SNP_pos'] == True]
    base_matches = mismatch_df.loc[(mismatch_df['SNP_base'] == True) & (mismatch_df['SNP_pos'] == True)]
    if save_dir is not None:
        mismatch_df.to_csv(os.path.join(save_dir, '%s_SNPs_thresh%spct.csv' % (mismatch_fname, int(pct_threshold*100))), index=False)
        #print(mismatch_df.to_string())
    return (pos_matches_unique, pos_matches, base_matches)

pos_matches_unique, pos_matches, base_matches = seam_snp_intersection(mismatch_df, snp_df, save_dir=save_dir)
observed_statistic = 100*(len(base_matches)/len(snp_df))

print('Percent of positions (with any nucleotide) where both a SEAM mismatch and SNP occur: %.2f%s' % (100*(pos_matches_unique/len(snp_df)), '%'))
print('Percent of matching SEAM/SNP positions with qualifying bases: %.2f%s' % (100*(len(base_matches)/len(pos_matches)), '%'))


if 1:
    FF = mismatch_df[(mismatch_df['SNP_pos'] == False) & (mismatch_df['SNP_base'] == False)].shape[0]
    TF = mismatch_df[(mismatch_df['SNP_pos'] == True) & (mismatch_df['SNP_base'] == False)].shape[0]
    TT = mismatch_df[(mismatch_df['SNP_pos'] == True) & (mismatch_df['SNP_base'] == True)].shape[0]
    TF_conflicts = mismatch_df[(mismatch_df['SNP_conflict'] == True)].shape[0]
    bars = plt.bar([0,1,2,3], [FF, TF, TF_conflicts, TT], color='lightgray', edgecolor='k', linewidth=3)
    plt.xticks(range(4), ['F/F', 'T/F', 'T/F Conflicts', 'T/T'])
    #bars[0].set_edgecolor('#8a2be2')
    #bars[1].set_edgecolor('magenta')
    #bars[2].set_edgecolor('magenta')
    plt.rcParams['savefig.dpi'] = 200
    plt.tight_layout()
    plt.show()
    print('T/F conflicts: %.3f%s' % ((TF_conflicts/TF)*100,'%'))
    print('')

#def mcnemar_test(mismatch_df):
#    from statsmodels.stats.contingency_tables import mcnemar
#    # perform McNemar test over matching SNP/SEAM positions, considering nucleotide content
#    '''contingency_table = pd.crosstab(mismatch_df['SNP_pos'], mismatch_df['SNP_base'])'''
#    a = mismatch_df[(mismatch_df['SNP_pos'] == True) & (mismatch_df['SNP_base'] == True)].shape[0]  # number of concordant pairs (both True)
#    b = mismatch_df[(mismatch_df['SNP_pos'] == True) & (mismatch_df['SNP_base'] == False)].shape[0]  # number of discordant pairs (A True, B False)
#    c = mismatch_df[(mismatch_df['SNP_pos'] == False) & (mismatch_df['SNP_base'] == True)].shape[0]  # number of discordant pairs (A False, B True)
#    d = mismatch_df[(mismatch_df['SNP_pos'] == False) & (mismatch_df['SNP_base'] == False)].shape[0]
#    print([[a, b], [c, d]])
#    result = mcnemar([[a, b], [c, d]], exact=True)
#    print("McNemar's test statistic:", result.statistic)
#    print("P-value:", result.pvalue)

#mcnemar_test(mismatch_df)


def binomial_test(mismatch_df):
    num_successes = mismatch_df[(mismatch_df['SNP_base'] == True)].shape[0]
    num_trials = len(mismatch_df)
    #probabilities = [0.6] * 4 + [0.3] * 6 + 1 # 4/10 inputs have 1/2 probability; 6/10 inputs have 2/10 probability
    #weighted_avg_prob = sum(probabilities) / len(probabilities)
    weighted_avg_prob = (0.3*6 + 0.6*4 + 1)/11.
    print('')
    #print('Weighted average probability:', weighted_avg_prob)
    p_value = stats.binomtest(num_successes, num_trials, p=weighted_avg_prob, alternative='less') #  probability of observing the given number of successes under H0 by chance
    print('Binomial test P-value:', p_value) # if sufficiently small, reject H0 and conclude that the observed number of successes is significantly less than expected by chance

binomial_test(mismatch_df)


if 0: # compute null distribution (scramble bases in SNP dataframe over many trials)
    snp_df_null = snp_df.copy(deep=True)
    allele_options = ['B','D','H','V']
    permuted_statistics = []
    num_trials = 10000
    print('')
    print('Computing null distribution...')
    for i in tqdm(range(num_trials), desc='Trials'):
        for j in range(len(snp_df_null)):
            if snp_df['allele'].iloc[j] == 'A':
                allele_options.extend('RMW')
            if snp_df['allele'].iloc[j] == 'C':
                allele_options.extend('YMS')
            if snp_df['allele'].iloc[j] == 'G':
                allele_options.extend('RKS')
            if snp_df['allele'].iloc[j] == 'T':
                allele_options.extend('YKW')
            snp_df_null.loc[j, 'allele'] = np.random.choice(allele_options, size=1)
        pos_matches_unique, pos_matches, base_matches = seam_snp_intersection(mismatch_df, snp_df_null, save_dir=None)
        permuted_statistics.append(100*(len(base_matches)/len(snp_df)))

    #norm_stat, norm_pval = stats.shapiro(permuted_statistics)
    norm_stat, norm_pval = stats.kstest(permuted_statistics, 'norm')

    if norm_pval >= 0.05: # z-test
        print('Fail to reject H0: not enough evidence to suggest sample does not come from normal distribution')
        mean_trials = np.mean(permuted_statistics)
        std_trials = np.std(permuted_statistics)
        print(mean_trials, std_trials)
        test_statistic = (observed_statistic - mean_trials) / (std_trials / np.sqrt(num_trials))
        p_value = 2 * (1 - stats.norm.cdf(np.abs(test_statistic))) # 2-tailed test for z-test
        print('Z-test P-value:', p_value)
    else:
        print('Reject H0. The data is not normally distributed.')
        U_statistic, p_value = stats.mannwhitneyu([observed_statistic], permuted_statistics, alternative='less') # greater: x is less than the distribution of y
        print('MWU P-value:', p_value)

    if 1: # permutation statistics by counts/proportion
        num_extreme = np.sum(np.array(permuted_statistics) <= observed_statistic)
        p_value = num_extreme / num_trials
        print('P-value (counting):', p_value)
        p_value = np.mean(np.array(permuted_statistics) <= observed_statistic)
        print('P-value (proportion):', p_value)

    if 1:
        plt.hist(permuted_statistics, bins=int(np.sqrt(num_trials)))
        plt.axvline(observed_statistic, color='k', label='observed')
        plt.legend()
        plt.tight_layout()
        plt.show()




        




import os, sys
sys.dont_write_bytecode = True
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import logomaker
import squid
py_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(py_dir)


if 1:
    dir_name = 'outputs_local_chrombpnet_PPIF_promoter/mut_pt10/seed2_N100k_allFolds'
    attr_map = 'deepshap'
    alphabet = ['A','C','G','T']


inputs_dir = os.path.join(parent_dir, 'examples/%s' % dir_name)
mave_fname = os.path.join(inputs_dir, 'mave.csv')
mave = pd.read_csv(mave_fname)
nS = len(mave)
seq_length = len(mave['Sequence'][0])
dim = (seq_length)*len(''.join(alphabet))
maps_fname = os.path.join(inputs_dir, 'maps_%s.npy' % attr_map)
maps = np.load(maps_fname, mmap_mode='r')


if 0:
    for i in range(0,50):
        fig, ax = plt.subplots(figsize=(10,1))
        logo = squid.utils.arr2pd(maps[i,278:385,:], alphabet)
        logomaker.Logo(df=logo,
                        ax=ax,
                        fade_below=.5,
                        shade_below=.5,
                        width=.9,
                        center_values=True,
                        font_name='Arial Rounded MT Bold',
                        )
        ax.axis('off')
        plt.rcParams['savefig.dpi'] = 200

        #plt.show()
        plt.savefig(os.path.join(py_dir, 'map_%s.png' % i), facecolor='w', dpi=200)
        plt.close()

if 1:
    for i in [0, 7, 17, 1, 42]:
        print(mave['Sequence'][i][278:385])


import os, sys
sys.dont_write_bytecode = True
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
import matplotlib.gridspec as gridspec
import matplotlib.patches as mpatches
from matplotlib.ticker import MaxNLocator
from matplotlib.figure import Figure
from mpl_toolkits.axes_grid1 import make_axes_locatable
import logomaker
import pandas as pd
import seaborn as sns
from scipy.spatial import distance
from scipy.cluster import hierarchy
import squid.utils as squid_utils # pip install squid-nn
from tqdm import tqdm
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)



def plot_logo(logo_df, logo_type, axis=None, aesthetic_lvl=1, ref_seq=None, fontsize=4, figsize=None, y_min_max=None, center_values=True):
    """Function to plot attribution maps or sequence statistics as logos.

    Parameters
    ----------
    logo_df : pandas.dataframe 
        Dataframe containing attribution map values or sequence statistics with proper column headings for alphabet.
        See function 'arr2pd' in utilities.py for conversion of array to Pandas dataframe.
    logo_type : str {'attribution', 'sequence'}
        Data type of input dataframe
    axis : matplotlib.pyplot axis
        Axis object for plotting logo
    aesthetic_lvl : int {1, 2, 3}
        Increasing this value increases the aesthetic appearance of the logo, with the caveat that higher
        aesthetic levels have higher computation times.
    ref_seq : str or None
        The reference sequence (e.g., wild type) for the ensemble of mutagenized sequences.
    fontsize : int
        The font size for the axis ticks.
    figsize : [float, float]
        The figure size, if no axis is provided.
    y_min_max : [float, float]
        The global attribution minimum and maximum value over all attribution maps (not just the current cluster).
    center_values : boole

    Returns
    -------
    logo : pandas.dataframe
    axis : matplotlib.pyplot axis
    """
    map_length = len(logo_df)
    color_scheme = 'classic'
    if logo_type == 'sequence' and ref_seq is not None:
        color_scheme = 'dimgray'
    
    if axis is None:
        fig, axis = plt.subplots()
        if figsize is not None:
            fig.set_size_inches(figsize) # e.g., [20, 2.5] or [10, 1.5] for logos < 100 nt

    if aesthetic_lvl > 0:
        if aesthetic_lvl == 2: # pretty logomaker mode (very slow)
            logo = logomaker.Logo(df=logo_df,
                                ax=axis,
                                fade_below=.5,
                                shade_below=.5,
                                width=.9,
                                center_values=center_values,
                                font_name='Arial Rounded MT Bold', # comp time bottleneck
                                color_scheme=color_scheme)
        elif aesthetic_lvl == 1: # plain logomaker mode (faster)
            if logo_type == 'sequence' and ref_seq is not None:
                logo = logomaker.Logo(df=logo_df,
                                    ax=axis,
                                    width=.9,
                                    center_values=center_values,
                                    color_scheme=color_scheme)
            else:
                logo = logomaker.Logo(df=logo_df,
                                    ax=axis,
                                    fade_below=.5,
                                    shade_below=.5,
                                    width=.9,
                                    center_values=center_values,
                                    color_scheme=color_scheme)
        if logo_type == 'sequence' and ref_seq is not None:
            logo.style_glyphs_in_sequence(sequence=ref_seq, color='darkorange')
        logo.style_spines(visible=False)
        logo.style_spines(spines=['left'], visible=True)
        logo.style_xticks(rotation=90, fmt='%d', anchor=0)
        logo.ax.xaxis.set_ticks_position('none')
        logo.ax.xaxis.set_tick_params(pad=-1)
        #logo.ax.yaxis.set_major_formatter(FormatStrFormatter('%.2f'))
    else:
        logo, axis = plot_logo_simple(np.array(logo_df), axis, color_scheme, ref_seq=ref_seq, center_values=center_values)
        axis.spines['top'].set_visible(False)
        axis.spines['right'].set_visible(False)
        axis.spines['bottom'].set_visible(False)
        axis.tick_params(axis='x', rotation=90)#, anchor=0)
        axis.xaxis.set_major_locator(MaxNLocator(integer=True))
        #axis.yaxis.set_major_formatter(FormatStrFormatter('%.2f'))
        axis.tick_params(bottom=False)
        axis.axhline(zorder=100, color='black', linewidth=0.5)

    axis.tick_params(axis="x", labelsize=fontsize)
    axis.tick_params(axis="y", labelsize=fontsize)

    axis.ticklabel_format(style='sci', axis='y', scilimits=(-3,3))
    axis.yaxis.offsetText.set_fontsize(fontsize)

    if center_values is True:
        centered_df = center_mean(np.array(logo_df))
        ymin = min(np.sum(np.where(centered_df<0,centered_df,0), axis=1))
        ymax = max(np.sum(np.where(centered_df>0,centered_df,0), axis=1))
    else:
        ymin = min(np.sum(np.where(np.array(logo_df)<0,np.array(logo_df),0), axis=1))
        ymax = max(np.sum(np.where(np.array(logo_df)>0,np.array(logo_df),0), axis=1))
    axis.set_ylim([ymin, ymax])

    if y_min_max is not None:
        if np.abs(y_min_max[0]) > np.abs(y_min_max[1]):
            y_max_abs = np.abs(y_min_max[0])
        else:
            y_max_abs = np.abs(y_min_max[1])
        plt.ylim(-1.*y_max_abs, y_max_abs)
    
    if map_length <= 100:
        axis.set_xticks(np.arange(0, map_length, 1))
    if map_length > 100:
        axis.set_xticks(np.arange(0, map_length-1, 5))
    if map_length > 1000:
        axis.set_xticks(np.arange(0, map_length-1, 10))

    plt.xlim(-0.5, map_length+.5)
    plt.tight_layout()
    return (logo, axis)



def plot_clusters_matches_sns(df, column='Entropy', sort=True, threshold=50, dpi=200, save_dir=None):
    """Function for visualizing all cluster positional statistics, including marginal distrubutions.

    Parameters
    ----------
    df : pandas.DataFrame, shape=(num_clusters x num_positions, 3)
        DataFrame corresponding to all mismatches (or matches) to reference sequence (or cluster consensus sequence).
        DataFrame is output by clusterer.py as 'compare_clusters.csv'
    column : str
        If 'Entropy', figure labels describe the Shannon entropy of characters at each position per cluster
        If 'Reference', figure labels describe the number of mismatches to reference sequence
        (recommended for local mutagenesis libraries).
        If 'Consensus', figure labels describe the number of matches to each cluster's respective consensus sequence
        (recommended for global mutagenesis libraries).
    sort : bool
        If True, the ordering of clusters will be sorted using the similarity of their positional elements.
    threshold : int in [0,100]
        Threshold value for whether to include positional elements in summation displayed in marginal plots.
        If 'reference' is True, all percent mismatches above the threshold will be summed.
        If 'reference' is False, all percent matches above the threshold will be summed,
        not including values at 100% to avoid obscuring the display due to presence of the conserved sequence pattern.
    dpi : int
        Controls the DPI used for rendering the figure.
    save_dir : str
        Directory for saving figures to file.
    """

    try:
        df = pd.read_csv(df)
    except:
        pass

    revels = df.pivot(columns='Position', index='Cluster', values=column)

    if sort is True: # reorder dataframe based on dendrogram sorting: 
        row_linkage = hierarchy.linkage(distance.pdist(revels), method='average')
        dendrogram = hierarchy.dendrogram(row_linkage, no_plot=True, color_threshold=-np.inf)
        reordered_ind = dendrogram['leaves']
        revels = df.pivot(columns='Position', index='Cluster', values=column)
        revels = revels.reindex(reordered_ind)

    nC = df['Cluster'].max() + 1
    nP = df['Position'].max() + 1

    # from https://stackoverflow.com/questions/65917235/how-to-create-a-heatmap-with-marginal-histograms-similar-to-a-jointplot
    joint = sns.jointplot(data=df, x='Position', y='Cluster', kind='hist', bins=(nC, nP))
    joint.ax_marg_y.cla()
    joint.ax_marg_x.cla()

    palette = sns.color_palette('rocket', n_colors=100)
    if column == 'Entropy':
        label = 'Shannon entropy (bits)'
    elif column == 'Reference':
        label = 'Mismatches to reference sequence'
        palette.reverse()
    elif column == 'Consensus':
        label = 'Matches to per-cluster consensus'

    heat = sns.heatmap(data=revels, ax=joint.ax_joint, cbar=False, cmap=palette,
                    cbar_kws={'format': '%.0f%%',
                                'label': label,
                                'location': 'bottom', #left
                                'orientation': 'horizontal'},
                                )

    # filter out all mismatches/matches based on threshold value (%):
    if column == 'Reference':
        df[column] = (df[column] >= threshold).astype(int) # binary mask
    elif column == 'Consensus':
        if threshold == 100:
            threshold_low = 99.9
        else:
            threshold_low = threshold
        df[column] = ((df[column] > threshold_low) & (df[column] < 100)).astype(int) # binary mask

    marg_color = 'dimgray'
    joint.ax_marg_y.barh(np.arange(0.5, nC), df.groupby(['Cluster'])[column].sum().to_numpy(), color=marg_color)
    joint.ax_marg_x.bar(np.arange(0.5, nP), df.groupby(['Position'])[column].sum().to_numpy(), color=marg_color)
    joint.ax_marg_x.yaxis.tick_right()

    if nP > 100:
        x_skip = 10
    else:
        x_skip = 1
    xtick_labels = []
    xtick_range = np.arange(0.5, nP, x_skip)
    for i in xtick_range:
        if int(i)%x_skip == 0:
            xtick_labels.append(str(int(i-0.5)))
    joint.ax_joint.set_xticks(xtick_range)
    joint.ax_joint.set_xticklabels(xtick_labels, rotation=0)

    if nC > 10:
        y_skip = 6
    else:
        y_skip = 1
    ytick_labels = []
    ytick_range = np.arange(0.5, nC, y_skip)
    y_idx = 0
    for i in ytick_range:
        if int(i)%y_skip == 0:
            if sort is True:
                ytick_labels.append(str(reordered_ind[y_idx]))
            else:
                ytick_labels.append(str(int(i-0.5)))
        y_idx += y_skip
    joint.ax_joint.set_yticks(ytick_range)
    joint.ax_joint.set_yticklabels(ytick_labels, rotation=0)

    joint.ax_joint.tick_params(axis='both', which='major', labelsize=6)
    joint.ax_marg_x.tick_params(axis='both', which='major', labelsize=6)
    joint.ax_marg_y.tick_params(axis='both', which='major', labelsize=6)
    joint.ax_marg_x.yaxis.set_major_locator(MaxNLocator(integer=True))
    joint.ax_marg_y.xaxis.set_major_locator(MaxNLocator(integer=True))

    # remove ticks between heatmap and histograms:
    joint.ax_marg_x.tick_params(axis='x', bottom=False, labelbottom=False)
    joint.ax_marg_y.tick_params(axis='y', left=False, labelleft=False)
    if 0: # remove ticks showing the heights of the histograms
        joint.ax_marg_x.tick_params(axis='y', left=False, labelleft=False)
        joint.ax_marg_y.tick_params(axis='x', bottom=False, labelbottom=False)
        joint.ax_marg_x.set_frame_on(False)
        joint.ax_marg_y.set_frame_on(False)

    # plot colorbar externally (hack):
    fig = plt.figure()
    gs = gridspec.GridSpec(1, 2, width_ratios=[30,1])
    mg0 = SeabornFig2Grid(joint, fig, gs[0])
    ax2 = plt.subplot(gs[1])
    ax2.get_figure().colorbar(mpl.cm.ScalarMappable(norm=plt.Normalize(0, 100), cmap=mpl.colors.ListedColormap(list(palette))),
                cax=ax2, orientation='vertical', label=label,
                format='%.0f%%')

    # plot marginal legend externally:
    if column == 'Reference':
        if threshold == 100:
            threshold_operator = '='
        else:
            threshold_operator = '>'
        marginal_patch = mpatches.Patch(color=marg_color, label=r'Marginals: mismatches %s %s%s' % (threshold_operator, threshold, '%'))
    elif column == 'Consensus':
        marginal_patch = mpatches.Patch(color=marg_color, label=r'Marginals: %s%s < matches < 100%s' % (threshold_low, '%', '%'))
    elif column == 'Entropy': #ZULU
        print('TBD')
    fig.legend(handles=[marginal_patch], frameon=False, prop={'size': 8}, loc='upper left')

    plt.rcParams['savefig.dpi'] = dpi
    plt.tight_layout()
    if save_dir is not None:
        if column == 'Reference':
            fig.savefig(os.path.join(save_dir, 'figure_clusters_mismatches.png'))
        elif column == 'Consensus':
            fig.savefig(os.path.join(save_dir, 'figure_clusters_matches.png'))
        elif column == 'Entropy':
            fig.savefig(os.path.join(save_dir, 'figure_clusters_entropy.png'))
    plt.show()


def plot_clusters_matches_2d_gui(df, figure, sort=None, column='Entropy', delta=False, mut_rate=10, alphabet=['A','C','G','T'], sort_index=None):
    """Function for visualizing all cluster positional statistics. Does not include marginal distrubutions.

    Parameters
    ----------
    df : pandas.DataFrame, shape=(num_clusters x num_positions, 3)
        DataFrame corresponding to all mismatches (or matches) to reference sequence (or cluster consensus sequence).
        DataFrame is output by clusterer.py as 'mismatches_reference.csv' or 'matches_consensus.csv'.
    figure : Matplotlib.Figure
        Instantiated Figure from SEAM GUI.
    sort : {None, 'visual', 'predefined'}
        If 'visual', the ordering of clusters will be sorted using the similarity of their sequence-derived patterns.
        If 'predefined', the ordering of clusters will be sorted using a predefined list (e.g., the median DNN score of each cluster)
    column : str
        If 'Entropy', figure labels describe the Shannon entropy of characters at each position per cluster.
        If 'Reference', figure labels describe the number of mismatches to reference sequence
        (recommended for local mutagenesis libraries).
        If 'Consensus', figure labels describe the number of matches to each cluster's respective consensus sequence
        (recommended for global mutagenesis libraries).
    delta : bool
        If True and column='Entropy', display matrix using the change in entropy from the background expectation.
    mut_rate : float
        Mutation rate (as a percentage) used to initially generate the mutagenized sequences. Only required if 'delta' is True.
    alphabet : list
        The alphabet used to determine the C characters in the logo such that
        each entry is a string; e.g., ['A','C','G','T'] for DNA.
    """
    sys.setrecursionlimit(100000) # fix: https://stackoverflow.com/questions/57401033/how-to-fixrecursionerror-maximum-recursion-depth-exceeded-while-getting-the-st

    ax = figure.add_subplot(111)

    nC = df['Cluster'].max() + 1
    nP = df['Position'].max() + 1
    revels = df.pivot(columns='Position', index='Cluster', values=column)

    if 0: # optional: set cells in MSM to be perfectly square
        ax.set_aspect('equal')

    if sort == 'visual': # reorder dataframe based on dendrogram
        row_linkage = hierarchy.linkage(distance.pdist(revels), method='average')
        dendrogram = hierarchy.dendrogram(row_linkage, no_plot=True, color_threshold=-np.inf)
        reordered_ind = dendrogram['leaves']
        revels = df.pivot(columns='Position', index='Cluster', values=column) # ZULU redundant?
        revels = revels.reindex(reordered_ind)
    elif sort == 'predefined':
        reordered_ind = sort_index
        revels = df.pivot(columns='Position', index='Cluster', values=column) # ZULU redundant
        revels = revels.reindex(reordered_ind)
        if nC != df['Cluster'].nunique():
            ax.set_aspect('equal')
            nC = df['Cluster'].nunique()
    elif sort is None:
        reordered_ind = np.arange(nC)

    if column == 'Entropy':
        palette = sns.color_palette('rocket', n_colors=100)
        if delta is False:
            label = 'Shannon entropy (bits)'
            palette.reverse()
        else:
            label = '$\Delta H$ (bits)'
            palette = sns.color_palette('vlag', n_colors=100)
    elif column == 'Reference':
        if 0:
            palette = sns.color_palette('rocket', n_colors=100)
        else:
            #palette = sns.color_palette('viridis', n_colors=100)
            palette = sns.color_palette('Blues', n_colors=100)
            palette.reverse()
        label = 'Mismatches to reference sequence'
        palette.reverse()
    elif column == 'Consensus':
        palette = sns.color_palette('rocket', n_colors=100)
        label = 'Matches to per-cluster consensus'

    if column == 'Entropy' and delta is True:
        from matplotlib import colors
        from scipy.stats import entropy
        r = mut_rate / 100
        p = np.array([1-r, r/float(len(alphabet)-1), r/float(len(alphabet)-1), r/float(len(alphabet)-1)])
        entropy_bg = entropy(p, base=2) # calculate Shannon entropy
        revels -= entropy_bg

        if 0: # remove noise based on threshold for cleaner visualization
            thresh = .10
            revels = revels.mask((revels < thresh) & (revels > -1.*thresh))

        divnorm = colors.TwoSlopeNorm(vmin=revels.min(numeric_only=True).min(), vcenter=0., vmax=revels.max(numeric_only=True).max())
        heatmap = ax.pcolormesh(revels, cmap='seismic', norm=divnorm)
    else:
        heatmap = ax.pcolormesh(revels, cmap=mpl.colors.ListedColormap(list(palette)))

    ax.set_xlabel('Position', fontsize=6)
    ax.set_ylabel('Cluster', fontsize=6)
    ax.tick_params(axis='both', which='major', labelsize=4)
    ax.set_title('Click on a cell to view its sequence variation', fontsize=5)
    ax.invert_yaxis()

    if nP > 100:
        x_skip = 10
    elif nP > 1000:
        x_skip = 20
    else:
        x_skip = 1
    xtick_labels = []
    xtick_range = np.arange(0.5, nP, x_skip)
    for i in xtick_range:
        if int(i)%x_skip == 0:
            xtick_labels.append(str(int(i-0.5)))
    ax.set_xticks(xtick_range)
    ax.set_xticklabels(xtick_labels, rotation=0, minor=False)

    if nC > 10:
        y_skip = 10
    else:
        y_skip = 1
    ytick_labels = []
    ytick_range = np.arange(0.5, nC, y_skip)
    y_idx = 0
    for i in ytick_range:
        if int(i)%y_skip == 0:
            if sort is True:
                ytick_labels.append(str(reordered_ind[y_idx]))
            else:
                ytick_labels.append(str(int(i-0.5)))
        y_idx += y_skip
    ax.set_yticks(ytick_range)
    ax.set_yticklabels(ytick_labels, rotation=0, minor=False)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    if column == 'Reference' or column == 'Consensus':
        cbar = figure.colorbar(heatmap, cax=cax, orientation='vertical', format='%.0f%%')
        heatmap.set_clim(0, 100)
    elif column == 'Entropy':
        cbar = figure.colorbar(heatmap, cax=cax, orientation='vertical')#, format='%.0f')
        if delta is False:
            heatmap.set_clim(0, 2)
    cbar.ax.set_ylabel(label, rotation=270, fontsize=6, labelpad=9)
    cbar.ax.tick_params(labelsize=6)

    return (ax, cax, reordered_ind, revels)


class SeabornFig2Grid():
    # https://stackoverflow.com/questions/69833665/seaborn-subplot-of-jointplots-doesnt-work
    def __init__(self, seaborngrid, fig,  subplot_spec):
        self.fig = fig
        self.sg = seaborngrid
        self.subplot = subplot_spec
        if isinstance(self.sg, sns.axisgrid.FacetGrid) or \
            isinstance(self.sg, sns.axisgrid.PairGrid):
            self._movegrid()
        elif isinstance(self.sg, sns.axisgrid.JointGrid):
            self._movejointgrid()
        self._finalize()

    def _movegrid(self):
        """ Move PairGrid or Facetgrid """
        self._resize()
        n = self.sg.axes.shape[0]
        m = self.sg.axes.shape[1]
        self.subgrid = gridspec.GridSpecFromSubplotSpec(n,m, subplot_spec=self.subplot)
        for i in range(n):
            for j in range(m):
                self._moveaxes(self.sg.axes[i,j], self.subgrid[i,j])

    def _movejointgrid(self):
        """ Move Jointgrid """
        h= self.sg.ax_joint.get_position().height
        h2= self.sg.ax_marg_x.get_position().height
        r = int(np.round(h/h2))
        self._resize()
        self.subgrid = gridspec.GridSpecFromSubplotSpec(r+1,r+1, subplot_spec=self.subplot)

        self._moveaxes(self.sg.ax_joint, self.subgrid[1:, :-1])
        self._moveaxes(self.sg.ax_marg_x, self.subgrid[0, :-1])
        self._moveaxes(self.sg.ax_marg_y, self.subgrid[1:, -1])

    def _moveaxes(self, ax, gs):
        # from https://stackoverflow.com/a/46906599/4124317
        ax.remove()
        ax.figure=self.fig
        self.fig.axes.append(ax)
        self.fig.add_axes(ax)
        ax._subplotspec = gs
        ax.set_position(gs.get_position(self.fig))
        ax.set_subplotspec(gs)

    def _finalize(self):
        plt.close(self.sg.fig)
        self.fig.canvas.mpl_connect("resize_event", self._resize)
        self.fig.canvas.draw()

    def _resize(self, evt=None):
        self.sg.fig.set_size_inches(self.fig.get_size_inches())


def center_mean(x): # equivalent to center_values in Logomaker
    mu = np.expand_dims(np.mean(x, axis=1), axis=1) #position dependent
    z = (x - mu)
    return z


def plot_logo_simple(logo_df, axis, color_scheme, ref_seq=None, center_values=True):
    # Adapted from:
        # https://github.com/kundajelab/deeplift/blob/16ef5dd05c3e05e9e5c7ec04d1d8a24fad046d96/deeplift/visualization/viz_sequence.py
        # https://github.com/kundajelab/chrombpnet/blob/a5c231fdf231bb29e9ca53d42a4c6e196f7546e8/chrombpnet/evaluation/figure_notebooks/subsampling/viz_sequence.py

    def ic_scale(pwm, background):
        odds_ratio = ((pwm+0.001)/(1.004))/(background[None,:])
        ic = ((np.log((pwm+0.001)/(1.004))/np.log(2))*pwm -\
                (np.log(background)*background/np.log(2))[None,:])
        return pwm*(np.sum(ic,axis=1)[:,None])

    def plot_a(ax, base, left_edge, height, color):
        a_polygon_coords = [
            np.array([
            [0.0, 0.0],
            [0.5, 1.0],
            [0.5, 0.8],
            [0.2, 0.0],
            ]),
            np.array([
            [1.0, 0.0],
            [0.5, 1.0],
            [0.5, 0.8],
            [0.8, 0.0],
            ]),
            np.array([
            [0.225, 0.45],
            [0.775, 0.45],
            [0.85, 0.3],
            [0.15, 0.3],
            ])
        ]
        for polygon_coords in a_polygon_coords:
            ax.add_patch(mpl.patches.Polygon((np.array([1,height])[None,:]*polygon_coords
                                                    + np.array([left_edge,base])[None,:]),
                                                    facecolor=color, edgecolor=color))

    def plot_c(ax, base, left_edge, height, color):
        ax.add_patch(mpl.patches.Ellipse(xy=[left_edge+0.65, base+0.5*height], width=1.3, height=height,
                                                facecolor=color, edgecolor=color))
        ax.add_patch(mpl.patches.Ellipse(xy=[left_edge+0.65, base+0.5*height], width=0.7*1.3, height=0.7*height,
                                                facecolor='white', edgecolor='white'))
        ax.add_patch(mpl.patches.Rectangle(xy=[left_edge+1, base], width=1.0, height=height,
                                                facecolor='white', edgecolor='white', fill=True))

    def plot_g(ax, base, left_edge, height, color):
        ax.add_patch(mpl.patches.Ellipse(xy=[left_edge+0.65, base+0.5*height], width=1.3, height=height,
                                                facecolor=color, edgecolor=color))
        ax.add_patch(mpl.patches.Ellipse(xy=[left_edge+0.65, base+0.5*height], width=0.7*1.3, height=0.7*height,
                                                facecolor='white', edgecolor='white'))
        ax.add_patch(mpl.patches.Rectangle(xy=[left_edge+1, base], width=1.0, height=height,
                                                facecolor='white', edgecolor='white', fill=True))
        ax.add_patch(mpl.patches.Rectangle(xy=[left_edge+0.825, base+0.085*height], width=0.174, height=0.415*height,
                                                facecolor=color, edgecolor=color, fill=True))
        ax.add_patch(mpl.patches.Rectangle(xy=[left_edge+0.625, base+0.35*height], width=0.374, height=0.15*height,
                                                facecolor=color, edgecolor=color, fill=True))

    def plot_t(ax, base, left_edge, height, color):
        ax.add_patch(mpl.patches.Rectangle(xy=[left_edge+0.4, base],
                    width=0.2, height=height, facecolor=color, edgecolor=color, fill=True))
        ax.add_patch(mpl.patches.Rectangle(xy=[left_edge, base+0.8*height],
                    width=1.0, height=0.2*height, facecolor=color, edgecolor=color, fill=True))

    default_colors = {0:'green', 1:'blue', 2:'orange', 3:'red'}
    default_plot_funcs = {0:plot_a, 1:plot_c, 2:plot_g, 3:plot_t}
    def plot_weights_given_ax(ax, array,
                    height_padding_factor,
                    length_padding,
                    subticks_frequency,
                    ref_index,
                    highlight,
                    colors=default_colors,
                    plot_funcs=default_plot_funcs,
                    ylabel="",
                    ylim=None):
        if len(array.shape)==3:
            array = np.squeeze(array)
        assert len(array.shape)==2, array.shape
        if (array.shape[0]==4 and array.shape[1] != 4):
            array = array.transpose(1,0)
        assert array.shape[1]==4
        max_pos_height = 0.0
        min_neg_height = 0.0
        heights_at_positions = []
        depths_at_positions = []
        # sort from smallest to highest magnitude
        for i in range(array.shape[0]):
            acgt_vals = sorted(enumerate(array[i,:]), key=lambda x: abs(x[1]))
            positive_height_so_far = 0.0
            negative_height_so_far = 0.0
            for letter in acgt_vals:
                plot_func = plot_funcs[letter[0]]
                if ref_index is not None and letter[0] == ref_index[i]:
                    color = 'darkorange'
                else:
                    color=colors[letter[0]]
                if (letter[1] > 0):
                    height_so_far = positive_height_so_far
                    positive_height_so_far += letter[1]
                else:
                    height_so_far = negative_height_so_far
                    negative_height_so_far += letter[1]
                plot_func(ax=ax, base=height_so_far, left_edge=i-0.5, height=letter[1], color=color)
            max_pos_height = max(max_pos_height, positive_height_so_far)
            min_neg_height = min(min_neg_height, negative_height_so_far)
            heights_at_positions.append(positive_height_so_far)
            depths_at_positions.append(negative_height_so_far)

        # now highlight any desired positions; the key of the highlight dict should be the color
        for color in highlight:
            for start_pos, end_pos in highlight[color]:
                assert start_pos >= 0.0 and end_pos <= array.shape[0]
                min_depth = np.min(depths_at_positions[start_pos:end_pos])
                max_height = np.max(heights_at_positions[start_pos:end_pos])
                ax.add_patch(
                    mpl.patches.Rectangle(xy=[start_pos,min_depth],
                        width=end_pos-start_pos,
                        height=max_height-min_depth,
                        edgecolor=color, fill=False))

        ax.set_xlim(-length_padding, array.shape[0]-length_padding)
        ax.xaxis.set_ticks(np.arange(0.0, array.shape[0]+1, subticks_frequency))

        # use user-specified y-axis limits
        if ylim is not None:
            min_neg_height, max_pos_height = ylim
            assert min_neg_height <= 0
            assert max_pos_height >= 0

        height_padding = max(abs(min_neg_height)*(height_padding_factor),
                            abs(max_pos_height)*(height_padding_factor))
        ax.set_ylim(min_neg_height-height_padding, max_pos_height+height_padding)
        ax.set_ylabel(ylabel)
        ax.yaxis.label.set_fontsize(15)

    if color_scheme == 'classic':
        colors = {0: [0, .5, 0], 1: [0, 0, 1], 2: [1, .65, 0], 3: [1, 0, 0]} # Logomaker 'classic' mode
    elif color_scheme == 'dimgray':
        colors = {0: 'dimgray', 1: 'dimgray', 2: 'dimgray', 3: 'dimgray'}
    plot_funcs = {0: plot_a, 1: plot_c, 2: plot_g, 3: plot_t}

    if ref_seq is not None:
        ref_index = []
        for c in ref_seq:
            if c == 'A':
                ref_index.append(0)
            elif c == 'C':
                ref_index.append(1)
            elif c == 'G':
                ref_index.append(2)
            elif c == 'T':
                ref_index.append(3)
    else:
        ref_index = None

    if center_values is True:
        logo_df = center_mean(logo_df)

    logo = plot_weights_given_ax(ax=axis, array=logo_df,
                                height_padding_factor=0.2,
                                length_padding=.5,#1.0,
                                subticks_frequency=100.0,
                                colors=colors, plot_funcs=plot_funcs,
                                ref_index=ref_index,
                                highlight={}, ylabel="")
    return (logo, axis)


def mismatch2csv(comparison_df, clusters, sequences, reference, threshold=100, sort_index=None, alphabet=['A','C','G','T'], verbose=1, save_dir=None):
    # input sequences need to be pre-sliced to match cropped attribution region (if applicable)

    nC = comparison_df['Cluster'].max() + 1
    nP = comparison_df['Position'].max() + 1

    if sort_index is not None:
        mapping_dict = {old_cluster: new_cluster for new_cluster, old_cluster in enumerate(sort_index)}
        comparison_df['Cluster'] = comparison_df['Cluster'].map(mapping_dict)
        clusters['Cluster'] = clusters['Cluster'].map(mapping_dict)

    comparison_df = comparison_df.drop(comparison_df[comparison_df['Reference'] < threshold].index)

    column_names = ['Position', 'Cluster', 'Reference', 'Pct_mismatch']
    for char in alphabet:
        column_names.append(char)
    mismatch_df = pd.DataFrame(columns=column_names)
    mismatch_df['Position'] = comparison_df['Position']
    mismatch_df['Cluster'] = comparison_df['Cluster']
    mismatch_df['Pct_mismatch'] = comparison_df['Reference']
    mismatch_df = mismatch_df.fillna(0)
    mismatch_df.reset_index(drop=True, inplace=True)   
     
    for i in range(len(mismatch_df)):
        row = int(mismatch_df['Cluster'].iloc[i])
        col = int(mismatch_df['Position'].iloc[i])
        mismatch_df.loc[i, 'Reference'] = reference[col]
        k_idxs = np.array(clusters['Cluster'].loc[clusters['Cluster'] == row].index)
        seqs = sequences['Sequence']
        seqs_cluster = seqs[k_idxs]
        occ = seqs_cluster.str.slice(col, col+1)
        vc = occ.value_counts()
        vc = vc.sort_index()
        chars = list(vc.index)
        vals = list(vc.values)
        for j in range(len(chars)):
            mismatch_df.loc[i, chars[j]] = vals[j]

    mismatch_df = mismatch_df.sort_values('Position')
    mismatch_df['Sum'] = mismatch_df[list(mismatch_df.columns[4:])].sum(axis=1)

    if save_dir is not None:
        mismatch_df.to_csv(os.path.join(save_dir, 'mismatches_%spct.csv' % (int(threshold))), index=False)
    else:
        if verbose == 1:
            print(mismatch_df.to_string())

    # print some statistics for better understanding the distribution:
    num_hits = len(mismatch_df)
    pos_unique = np.unique(mismatch_df['Position'].values)
    pos_intersect = np.intersect1d(pos_unique, np.arange(0,nP))
    cluster_unique = np.unique(mismatch_df['Cluster'].values)
    cluster_intersect = np.intersect1d(cluster_unique, np.arange(0,nC))
    if verbose == 1:
        print('\tTotal number of hits (i.e., %s%s mismatches): %s of %s (%.2f%s)' % (threshold, '%', num_hits, nP*nC, 100*(num_hits/(nP*nC)), '%'))
        print('\tPercent of positions with at least one hit: %.2f%s' % (100*(len(pos_intersect)/nP),'%'))
        print('\tPercent of clusters with at least one hit: %.2f%s' % (100*(len(cluster_intersect)/nC),'%'))
        print('\tAverage occupancy of target clusters:', mismatch_df['Sum'].mean())

    return mismatch_df



if __name__ == "__main__":
    py_dir = os.path.dirname(os.path.abspath(__file__))
    parent_dir = os.path.dirname(py_dir)

    #name = 'PIK3R3'#'CD40'#'NXF1'
    #mode = 'quantity'
    #mode = 'profile'

    #dir_name = 'examples/examples_clipnet/outputs_local_%s/heterozygous_pt01/%s_nfolds9' % (name, mode)
    dir_name = 'examples/examples_chrombpnet/outputs_local_PPIF_promoter_counts/mut_pt10/seed2_N100k_allFolds'
    #clusters_dir = '/clusters_umap_kmeans200'
    #clusters_dir = '/clusters_umap_dbscan'
    clusters_dir = '/clusters_hierarchical_maxclust200_sortMedian'
    #clusters_dir = '/clusters_hierarchical_cut0.0001'
    save_dir = os.path.join(parent_dir, dir_name + clusters_dir)
    comparison_df = pd.read_csv(os.path.join(parent_dir, dir_name + clusters_dir + '/compare_clusters.csv'))
    clusters = pd.read_csv(os.path.join(parent_dir, dir_name + clusters_dir + '/all_clusters.csv'))
    mave = pd.read_csv(os.path.join(parent_dir, dir_name + '/mave.csv'))
    reference = mave['Sequence'][0]
    mismatch_df = mismatch2csv(comparison_df, clusters, mave, reference, 
                               threshold=90,#100,
                               alphabet=['A','C','G','T','M','R','W','S','Y','K'], save_dir=save_dir)


from setuptools import setup, find_packages

setup(
    name="seam-nn",
    version="0.1.1",
    author="Evan Seitz",
    author_email="evan.e.seitz@gmail.com",
    packages=find_packages(),
    description = "SEAM is a Python package to interpret sequence-based deep learning models for regulatory genomics data using meta-explanations.",
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires=">=3.7.2",
    install_requires=[
    	'numpy',
	'matplotlib>=3.2.0',
	'pandas',
        'tqdm',
        'logomaker',
        'pyqt5',
        'psutil',
        'biopython',
        'tensorflow'
    ],
)


import os, sys, time
sys.dont_write_bytecode = True
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.spatial.distance import squareform
from scipy.cluster import hierarchy
from tqdm import tqdm
py_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(py_dir)


# =============================================================================
# User parameters
# =============================================================================

# to do: add umap, pca, etc.

gpu = True

attr_map = 'deepshap'
alphabet = ['A','C','G','T']
#data_dir = os.path.join(py_dir, 'outputs_AP1_mut0_seq13748') # location of required data (see below)
#dir_name = 'examples_deepmel2/outputs_REVO-idx45-class16-tw4-wl5101520-tm20-tr3'
dir_name = 'examples_chrombpnet/outputs_local_PPIF_promoter_counts/mut_pt01/N100k_fold0'
data_dir = os.path.join(parent_dir, 'examples/%s' % dir_name)
dist_fname = os.path.join(data_dir, 'distances.dat')
batch_size = 10000
link_method = 'ward' # linkage method for hierarchical clustering

# =============================================================================
# Load in required data
# =============================================================================

mave_fname = os.path.join(data_dir, 'mave.csv')
mave = pd.read_csv(mave_fname)
nS = len(mave)
seq_length = len(mave['Sequence'][0]) # or just set manually
dim = (seq_length)*len(''.join(alphabet))
maps_fname = os.path.join(data_dir, 'maps_%s.npy' % attr_map)
maps = np.load(maps_fname, mmap_mode='r')
maps = maps.reshape((nS, seq_length*len(alphabet)))

# =============================================================================
# Calculate distance matrix
# =============================================================================
if 1: # if distances not previously calculated
    print('Computing distances...')

    if gpu is False:
        t1 = time.time()
        # compute distances:
        D_upper = np.zeros(shape=(nS, nS))
        p = 2. # Minkowski distance metric: p1=Manhattan, p2=Euclidean, etc.
        for i in tqdm(range(0, nS), desc='Index'):
            for j in range(0, nS):
                if i < j:
                    D_upper[i,j] = np.linalg.norm(maps[i,:]-maps[j,:])
        D = D_upper + D_upper.T - np.diag(np.diag(D_upper))
        D_flat = squareform(D)
        D_flat_CPU = D_flat # for comparison below only
        t2 = time.time()
        print('Distances time:', t2-t1)

    elif gpu is True:
        import tensorflow as tf

        def distance_batch(A, B, batch_size=4096, output_file='distances.dat'):
            A = tf.cast(A, tf.float32)
            B = tf.cast(B, tf.float32)
            num_A = A.shape[0]
            num_B = B.shape[0]
            distance_matrix = np.memmap(output_file, dtype=np.float32, mode='w+', shape=(num_A, num_B))

            def pairwise_distance(A_batch, B_batch):
                v_A = tf.expand_dims(tf.reduce_sum(tf.square(A_batch), 1), 1)
                v_B = tf.expand_dims(tf.reduce_sum(tf.square(B_batch), 1), 1)
                p1 = tf.reshape(v_A, (-1, 1))
                p2 = tf.reshape(v_B, (1, -1))
                dist_squared = tf.add(p1, p2) - 2 * tf.matmul(A_batch, B_batch, transpose_b=True)
                dist_squared = tf.maximum(dist_squared, 0.0) # ensure non-negative values before taking sqrt
                return tf.sqrt(dist_squared)

            for i in tqdm(range(0, num_A, batch_size), desc='Distance batch'):
                A_batch = A[i:i + batch_size]
                for j in range(0, num_B, batch_size):
                    B_batch = B[j:j + batch_size]
                    distances = pairwise_distance(A_batch, B_batch).numpy()
                    distance_matrix[i:i + batch_size, j:j + batch_size] = distances # save intermediate results to memmap

            distance_matrix.flush() # flush memmap to ensure all data is written to disk

            return distance_matrix

        
        t1 = time.time()
        D = distance_batch(tf.Variable(maps), tf.Variable(maps), batch_size=batch_size, output_file=dist_fname)
        D = np.array(D).astype(np.float32)
        np.fill_diagonal(D, 0)
        D = (D + D.T) / 2 # matrix may be antisymmetric due to precision errors
        D_flat = squareform(D)
        os.remove(dist_fname)
        t2 = time.time()
        print('Distances time:', t2-t1)

    #print(np.allclose(D_flat_CPU, D_flat, atol=1e-8)) # sanity check

    if 0:
        np.save(os.path.join(data_dir, 'dist_upper_flat.npy'), D_flat)

else:
    D_flat = np.load(os.path.join(data_dir, 'dist_upper_flat.npy'))

# =============================================================================
# Hierarchical clustering on distance matrix
# =============================================================================
if 1:
    print('Computing hierarchical clusters...')
    t1 = time.time()
    linkage = hierarchy.linkage(D_flat, method=link_method, metric='euclidean')
    np.save(os.path.join(data_dir, 'hierarchical_linkage_%s_%s.npy' % (link_method, attr_map)), linkage)
    t2 = time.time()
    print('Linkage time:', t2-t1)


#if 0: # correlate the pairwise distances of all samples to those implied by the hierarchical clustering
#    from scipy.spatial import distance
#    c, coph_dists = hierarchy.cophenet(linkage, distance.pdist(maps))
#    print(c)



import os, sys
sys.dont_write_bytecode = True
import time
import numpy as np
import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.ticker import MaxNLocator
import warnings
warnings.filterwarnings("ignore", category=matplotlib.MatplotlibDeprecationWarning)
from matplotlib.ticker import FormatStrFormatter
from Bio import motifs # 'pip install biopython'
import logomaker
import squid.utils as squid_utils # 'pip install squid-nn'
import seaborn as sns # pip install seaborn
from scipy.stats import entropy
from tqdm import tqdm
import gc
py_dir = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.dirname(py_dir)

try:
    from . import utilities
except ImportError:
    import utilities

try:
    from . import impress
except ImportError:
    import impress

# to do items:
    # move 'generate_regulatory_df' into impress.py

# ZULU: clean up below function and the duplicate in _null.py and send both to utils.py
def generate_regulatory_df(mave, clusters_labels, clusters_idx, ref_seq): 
    # mave should already have sequence delimited
    mave['Cluster'] = clusters_labels
    seqs_all = mave['Sequence']
    seq_length = len(mave['Sequence'][0])
    regulatory_df = pd.DataFrame(columns = ['Cluster', 'Position', 'Reference', 'Consensus', 'Entropy'], index=range(len(clusters_idx)*seq_length))
    occ_idx = 0
    for k in tqdm(clusters_idx, desc='Clusters'):
        k_idxs =  mave.loc[mave['Cluster'] == k].index
        seqs_k = seqs_all[k_idxs]
        seq_array_cluster = motifs.create(seqs_k, alphabet=alphabet)
        pfm_cluster = seq_array_cluster.counts # position frequency matrix
        consensus_seq = pfm_cluster.consensus
        consensus_oh = squid_utils.seq2oh(consensus_seq, alphabet)
        # calculate matches to reference sequence:
        if ref_seq is not None:
            pos_occ1 = np.diagonal(np.eye(len(ref_seq))*ref_oh.dot(np.array(pd.DataFrame(pfm_cluster)).T))
            pos_occ1 = 100 - (pos_occ1/len(k_idxs))*100
        # calculate matches to cluster's consensus sequence:
        pos_occ2 = np.diagonal(np.eye(len(consensus_seq))*consensus_oh.dot(np.array(pd.DataFrame(pfm_cluster)).T))
        pos_occ2 = (pos_occ2/len(k_idxs))*100
        for pos in range(seq_length):
            occ = seqs_k.str.slice(pos, pos+1)
            vc = occ.value_counts(normalize=True).sort_index()
            regulatory_df.at[occ_idx, 'Cluster'] = k
            regulatory_df.at[occ_idx, 'Position'] = pos
            regulatory_df.at[occ_idx, 'Entropy'] = entropy(np.array(vc), base=2) # calculate Shannon entropy
            if ref_seq is not None:
                regulatory_df.at[occ_idx, 'Reference'] = pos_occ1[pos]
            else:
                regulatory_df.at[occ_idx, 'Reference'] = pd.NA
            regulatory_df.at[occ_idx, 'Consensus'] = pos_occ2[pos]
            occ_idx += 1
    return regulatory_df

# defaults
ref_idx = 0
plot_profiles = False
profiles_ylim = None

if 1:
    #dir_name = 'examples_deepstarr/outputs_local_AP1-m1-seq22612'
    #dir_name = 'examples_deepstarr/outputs_local_Ohler1-m0-seq20647'
    dir_name = 'examples_deepstarr/outputs_local_AP1-m0-seq13748/archives/AP1-rank0_origVersion'
    #dir_name = 'examples_deepstarr/outputs_local_Ohler1-m0-seq22627'
    #dir_name = 'examples_deepstarr/outputs_local_Ohler6-m0-seq171'
    #dir_name = 'examples_deepstarr/outputs_local_DRE-m2-seq4071'
    #dir_name = 'examples_deepstarr/outputs_local_AP1-m1-seq21069'
    #dir_name = 'examples_deepstarr/outputs_local_Ohler1-m0-seq20647/_other_librarySizes/librarySize_150k'
    ref_seq = 'First row'
    seq_stop = 249
    #attr_map = 'ism'
    #attr_map = 'saliency'
    #attr_map = 'smoothgrad'
    #attr_map = 'intgrad'
    attr_map = 'deepshap'
    #map_length = 249
elif 0:
    #dir_name = 'examples_deepstarr/outputs_global_AP1-N'
    dir_name = 'examples_deepstarr/outputs_global_CREB-NN'
    ref_seq = None
    seq_stop = 249
    attr_map = 'deepshap'
elif 0:
    dir_name = 'outputs_local_spliceAI_U2SURP/max'
    # U2SURP:
    ref_seq = 'GGGGGGACTACTGTATTATAAAAACTAATAATTGTCTTCTTTTCTCCCCTTTAAGTGGACGCGCCTTCAAGAAGAAATAGATCATCTGGTGGTAATACAGTTTTTTGCTCTTTTAATCGATAAATTT'
    seq_stop = 127
    attr_map = 'saliency'
elif 0:
    dir_name = 'examples_enformer/outputs_local_DNase_PPIF_promoter/seed2'
    #dir_name = 'examples_enformer/outputs_local_CAGE_PPIF_promoter/seed2'
    ref_seq = 'First row'
    attr_map = 'saliency'
    #map_length = 1024 # enhancer
    #map_length = 768 # promoter
elif 0:
    #dir_name = 'examples_chrombpnet/outputs_local_PPIF_promoter_counts/mut_pt10/seed2_N100k_allFolds'#fold4'
    #dir_name = 'examples_chrombpnet/outputs_local_PPIF_promoter_counts/mut_pt10/seed1and2_N200k_allFolds'
    dir_name = 'examples_chrombpnet/outputs_local_PPIF_promoter_counts/mut_pt40/N100k_fold0'
    ref_seq = 'First row'
    attr_map = 'deepshap'
    #map_length = 2048
elif 0:
    dir_name = 'outputs_local_enformer_DNase_PPIF_promoter/seed5_mutatePro_targetPro/promoter'
    ref_seq = 'First row'
    attr_map = 'saliency'
    #map_length = 768
elif 0:
    mode = 'quantity'
    #mode = 'profile' # {note for clipnet: cut=0.0001, 0.0002, 0.0004, 0.0008, 0.001, 0.003, 0.007}
    dir_name = 'examples_clipnet/outputs_local_PIK3R3/heterozygous_pt01/%s_nfolds9' % mode
    ref_seq = 'First row'
    attr_map = 'deepshap'
    #map_length = 1000
elif 0:
    #dir_name = 'examples_tfbind/outputs_8mer_SIX6'
    dir_name = 'examples_tfbind/outputs_8mer_Hnf4a' # 43728: GTAAACA
    #dir_name = 'examples_tfbind/outputs_8mer_Foxa2' # 11268: AGTAAACA
    #dir_name = 'examples_tfbind/outputs_8mer_Zfp187'
    #dir_name = 'examples_tfbind/outputs_8mer_Rfx3'
    #dir_name = 'examples_tfbind/outputs_8mer_Oct-1'
    ref_seq = 11268
    attr_map = 'ism'
    #map_length = 8
elif 0:
    tm = 20 # total mutations per EFS run
    #dir_name = 'examples_deepmel2/outputs_EFS-idx22-class16-traj0'
    #dir_name = 'examples_deepmel2/outputs_EFS-idx22-class16-traj20'
    #dir_name = 'examples_deepmel2/outputs_EFS-idx17-class16-traj20'
    #dir_name = 'examples_deepmel2/outputs_BEAM-idx45-class16-width2-muts16'
    #dir_name = 'examples_deepmel2/outputs_REVO-idx22-class16-tw3-wl20-tm20-tr5'
    dir_name = 'examples_deepmel2/outputs_REVO-idx45-class16-tw4-wl5101520-tm%s-tr3' % tm
    ref_seq = 'First row'
    attr_map = 'deepshap'
    #map_length = 500
    ref_idx = tm+1
elif 0:
    mode = 'profile'
    #mode = 'counts'
    dir_name = 'examples_procapnet/outputs_MYC_%s_fold0_r10' % mode
    ref_seq = 'First row'
    attr_map = 'deepshap'
    #map_length = 500
    profiles_ylim = [-769.2708160400391, 833.5737335205079] # profiles ylim over all seqs in library

aesthetic_lvl = 1 # {0, 1, 2} : modulates the aesthetic appearance of the logo, with the caveat that higher aesthetic levels have higher computation times

if 0:
    embedding = 'umap'#'pca'#'umap'
    if 0:
        clusterer = 'dbscan'
        n_clusters = None
    elif 1:
        clusterer = 'kmeans'
        n_clusters = 200
else:
    embedding = None
    clusterer = 'hierarchical'
    link_method = 'ward'
    if 0:
        cut_criterion = 'distance'
        cut_value = 8#10 # choose max cutoff distance for dendrogram
    else:
        cut_criterion = 'maxclust'
        cut_value = 200#100#500 # choose the desired number of clusters (same use as 'n_clusters' above)
    plot_dendro = True#False
    n_clusters = None # unused (see 'cut_value' above)

    
attr_logo_fixed = True
attr_logo_adaptive = True
seq_logo_pwm = False
seq_logo_enrich = False#True
plot_profiles = False # plot overlay of profiles associated with each cluster; requires file 'y_profiles.npy' corresponding to predictions for sequences in 'mave.csv' (both in same directory)
plot_bias = False#False # plot the per-cluster bias logos in the 'clusters_bias' folder
sort_median = True # sort the boxplots and regulatory matrix by the median DNN score of each cluster
bias_removal = False # removes bias based on subtracting the average of background attribution maps from each cluster 
mut_rate = .10 #.01 # mutation rate used for generating sequence library (if 'bias_removal' = True)
subset_clusters = False # focus on a specific subset of clusters (selected after first viewing all clusters; see below)
# ZULU: currently, sort_median must be false if subset_clusters is True

#sort_fig = 'visual'


inputs_dir = os.path.join(parent_dir, 'examples/%s' % dir_name)
if ref_seq == 'First row':
    ref_seq = pd.read_csv(os.path.join(inputs_dir, 'mave.csv'))['Sequence'][0]
    seq_stop = len(ref_seq)
    print('Sequence length:', seq_stop)
elif isinstance(ref_seq, int):
    ref_seq = pd.read_csv(os.path.join(inputs_dir, 'mave.csv'))['Sequence'][ref_seq]
    seq_stop = len(ref_seq)
    print('Sequence length:', seq_stop)


dpi = 200
eig1 = 0
eig2 = 1


def normalize(_d, to_sum=False, copy=True): # normalize all eigenvectors
    d = _d if not copy else np.copy(_d) # d is an (n x dimension) array
    d -= np.min(d, axis=0)
    d /= (np.sum(d, axis=0) if to_sum else np.ptp(d, axis=0)) # normalize to [0,1]
    return d


mave_fname = os.path.join(inputs_dir, 'mave.csv')
mave = pd.read_csv(mave_fname)
nS = len(mave)
maps_fname = os.path.join(inputs_dir, 'maps_%s.npy' % attr_map)

'''if 0: # only needed if column names in mave.csv are inconsistent with those used here
    mave.rename(columns={'y_DNN': 'DNN', 'x': 'Sequence'}, inplace=True)
    mave.to_csv(os.path.join(inputs_dir, 'mave_new.csv'), index=False)
    print('')
    print("Columns in 'mave.csv' renamed and saved as 'mave_new.csv'. To replace the old version, manually overwrite 'mave.csv' with 'mave_new.csv', then disable this code before rerunning.")
    print('')
    exit()'''

#alphabet = ['A','C','G','T']
alphabet = sorted(list(set(mave['Sequence'][0:100].apply(list).sum())))


if embedding is not None:
    output_name = '%s_%s_%s' % (attr_map, embedding, clusterer)
else:
    if clusterer == 'hierarchical':
        output_name = '%s_%s_%s%s' % (attr_map, clusterer, cut_criterion, cut_value)
    else:
        output_name = '%s_%s' % (attr_map, clusterer)
if n_clusters is not None:
    output_name = output_name + str(int(n_clusters))


if 1: # use cropped attribution maps
    # PPIF promoter:#seq_start = 186 #seq_stop = 406 # PPIF enhancer:
    if 1: # AP1-rank0 (DeepSTARR)
        seq_start = 110
        seq_stop = 137
    elif 0: # Global (DeepSTARR)
        seq_start = 114
        seq_stop = 143#2#1#3 DRE, AP1-N, CREB-NN
    elif 0: # PPIF promoter (ChromBPNet)
        seq_start = 190
        seq_stop = 690


    manifold_fname = os.path.join(inputs_dir, 'manifold_%s_%s_crop_%s_%s.npy' % (attr_map, embedding, seq_start, seq_stop))
    if sort_median is True:
        outputs_dir = os.path.join(inputs_dir, 'clusters_%s_crop_%s_%s_sortMedian' % (output_name, seq_start, seq_stop))
    else:
        outputs_dir = os.path.join(inputs_dir, 'clusters_%s_crop_%s_%s' % (output_name, seq_start, seq_stop))
    fontsize = 8 # for 30-nt (cropped) sequence
    map_length = seq_stop - seq_start
else: # use full attribution maps
    manifold_fname = os.path.join(inputs_dir, 'manifold_%s_%s.npy' % (attr_map, embedding))
    seq_start = 0

    if sort_median is True:
        outputs_dir = os.path.join(inputs_dir, 'clusters_%s_sortMedian' % output_name)
    else:
        outputs_dir = os.path.join(inputs_dir, 'clusters_%s' % output_name)
    fontsize = 4 # for 249-nt sequence


if not os.path.exists(outputs_dir):
    os.mkdir(outputs_dir)

seq_length = len(mave['Sequence'][0])
dim = (seq_length)*len(''.join(alphabet))
if 'map_length' in locals():
    pass
else:
    map_length = seq_length
print('Attribution length:', map_length)

if map_length < 10:
    figsize = [3, 1.5]
elif 10 <= map_length < 50:
    figsize = [10, 2.5]
elif 50 <= map_length < 250:
    figsize = [20, 2.5]
else:
    figsize = [20, 1.5]

if embedding is not None:
    manifold = np.load(manifold_fname)
    manifold = normalize(manifold)

try:
    maps = np.load(maps_fname)
    if maps.ndim == 2: # ZULU, consistency
        maps = maps.reshape((nS, seq_length, len(''.join(alphabet))))
except: # try to load numpy mmemap
    maps = np.memmap(maps_fname, dtype='float32', mode='r', shape=(nS, seq_length, len(''.join(alphabet))))


seq_length = seq_stop - seq_start
try:
    if seq_start != 0 and seq_stop != len(mave['Sequence'][0]):
        mave['Sequence'].str.slice(seq_start, seq_stop)
        maps = maps[:,seq_start:seq_stop,:]
        if ref_seq is not None:
            ref_seq = ref_seq[seq_start:seq_stop]
            ref_oh = squid_utils.seq2oh(ref_seq, alphabet)
except:
    pass


print('Clustering data...')
if embedding is not None:
    if clusterer == 'kmeans':
        from sklearn.cluster import KMeans # pip install -U scikit-learn
        clusters = KMeans(init='k-means++', n_clusters=n_clusters, random_state=0, n_init=10)
        clusters.fit(manifold)
        clusters_labels = clusters.labels_ # membership of each element to a given cluster
    elif clusterer == 'dbscan':
        from sklearn.cluster import DBSCAN
        clusters = DBSCAN(eps=0.01, min_samples=10).fit(manifold)
        clusters_labels = clusters.labels_ # membership of each element to a given cluster

    clusters_idx = np.arange(max(clusters_labels)+1)

    if 1:
        plt.scatter(manifold[:,eig1], manifold[:,eig2], c=clusters_labels, cmap='tab10', #edgecolor='none',
                    s=2.5, linewidth=.1, edgecolors='k')
        #plt.colorbar()
        plt.tight_layout()
        plt.savefig(os.path.join(outputs_dir, 'all_clusters.png'), facecolor='w', dpi=600)
        plt.close()

    all_clusters_df = pd.DataFrame(columns = ['Cluster', 'Psi1', 'Psi2'], index=range(nS))
    all_clusters_df['Cluster'] = clusters_labels
    mave['Cluster'] = clusters_labels
    all_clusters_df['Psi1'].iat[0] = eig1
    all_clusters_df['Psi2'].iat[0] = eig2
    sort_fig = 'visual'


else: # cluster attribution maps without embedding them
    if clusterer == 'hierarchical':
        from scipy.cluster import hierarchy
        linkage = np.load(os.path.join(inputs_dir, 'hierarchical_linkage_%s_%s.npy' % (link_method,attr_map)))

        if plot_dendro is True: # plot dendrogram
            sys.setrecursionlimit(100000) # fix: https://stackoverflow.com/questions/57401033/how-to-fixrecursionerror-maximum-recursion-depth-exceeded-while-getting-the-st
            plt.figure(figsize=(15, 10)) # (25,10)
            plt.title('Hierarchical Clustering Dendrogram')
            plt.xlabel('sample index')
            plt.ylabel('distance')
            with plt.rc_context({'lines.linewidth': 2}):
                hierarchy.dendrogram(linkage,
                                    leaf_rotation=90., # rotates the x-axis labels
                                    leaf_font_size=8., # font size for the x-axis labels
                                    )
            plt.xticks([])
            plt.gca().spines['top'].set_visible(False)
            plt.gca().spines['right'].set_visible(False)
            plt.xlabel('Clusters') 
            plt.savefig(os.path.join(os.path.dirname(outputs_dir), 'dendrogram.png'), facecolor='w', dpi=dpi)
            plt.close()
        
        clusters_labels = hierarchy.fcluster(linkage, cut_value, criterion=cut_criterion) # membership of each element to a given cluster
        clusters_labels -= 1 # zero-index clusters to match other methods

        '''if cut_criterion == 'maxclust': # optionally, print the cut level (the height at which the dendrogram is cut)
            max_d = 0
            for i in range(1, len(linkage) + 1):
                if len(np.unique(hierarchy.fcluster(linkage, i, criterion='maxclust'))) == cut_value:
                    max_d = linkage[i-1, 2]  # The height (or distance) at the last merge to achieve the desired clusters
                    break
            print("Cut level for desired number of clusters:", max_d)'''
        
        clusters_idx = np.unique(clusters_labels) #np.arange(max(clusters_labels)) #
        all_clusters_df = pd.DataFrame(columns = ['Cluster'], index=range(nS))
        all_clusters_df['Cluster'] = clusters_labels
        mave['Cluster'] = clusters_labels
        sort_fig = None


if subset_clusters is True: # option to only focus on a subselection of clusters (default=0)
    ref_cluster = all_clusters_df.loc[ref_idx, 'Cluster']
    #cluster_sorted_indices = [ref_cluster, 71, 147, 96, 91, 50, 79, 0, 160, 93, 32, 80, 45, 131, 179, 55] # circle
    cluster_sorted_indices = [0,160,93,32,80,45,131,179,55,ref_cluster]
    ref_cluster = 9#0 # reset to match updated position in list above
    all_clusters_df = all_clusters_df[all_clusters_df['Cluster'].isin(cluster_sorted_indices)]
    mave = mave[mave['Cluster'].isin(cluster_sorted_indices)]
    clusters_idx = [item for item in clusters_idx if item in cluster_sorted_indices]
    clusters_labels = [item for item in clusters_labels if item in cluster_sorted_indices]
    # create a dictionary to map the original clusters to their indices in cluster_sorted_indices
    cluster_index_map = {cluster: index for index, cluster in enumerate(cluster_sorted_indices)}
    # sort clusters_idx to match the order in cluster_sorted_indices
    clusters_idx = sorted(clusters_idx, key=lambda x: cluster_index_map[x])
    # update 'all_clusters.csv' with new column containing sorted indices (for user reference only)
    mapping_dict = {old_cluster: new_cluster for new_cluster, old_cluster in enumerate(cluster_sorted_indices)}
    all_clusters_df['Cluster_sort'] = all_clusters_df['Cluster'].map(mapping_dict)
    sort_median = False
    sort_fig = 'predefined'


all_clusters_df.to_csv(os.path.join(outputs_dir, 'all_clusters.csv'), index=False)

mave['Sequence'] = mave['Sequence'].str.slice(seq_start, seq_stop)
#seqs_all = mave['Sequence'].str.slice(seq_start, seq_stop)
seq_array_background = motifs.create(mave['Sequence'], alphabet=alphabet)#seqs_all)
pfm_background = seq_array_background.counts

if 1: # render boxplots for all clusters
    print('Rendering histograms and boxplots...')

    boxplot_data = []
    for k in tqdm(clusters_idx, desc='Clusters'):
        if k >= 0: # this line is a placeholder for optionally inserting code for a specific cluster
            k_idxs =  mave.loc[mave['Cluster'] == k].index
            #if len(k_idxs) >= 50: #optional threshold to handle oversplitting
            boxplot_data.append(mave['DNN'][k_idxs])

    # calculate the average IQR
    iqr_values = [np.percentile(data, 75) - np.percentile(data, 25) for data in boxplot_data]
    average_iqr = np.mean(iqr_values)

    if sort_median is True: # sort boxplots by ascending median
        cluster_medians = [np.median(sublist) for sublist in boxplot_data] # calculate the median for each boxplot
        cluster_sorted_indices = sorted(range(len(cluster_medians)), key=lambda i: cluster_medians[i]) # get the indices of the boxplots sorted based on the median values
        boxplot_data = [boxplot_data[i] for i in cluster_sorted_indices]
        clusters_idx = [clusters_idx[i] for i in cluster_sorted_indices]
        sort_fig = 'predefined'
        # update 'all_clusters.csv' with new column containing sorted indices (for user reference only)
        mapping_dict = {old_cluster: new_cluster for new_cluster, old_cluster in enumerate(cluster_sorted_indices)}
        all_clusters_df['Cluster_sort'] = all_clusters_df['Cluster'].map(mapping_dict)
        all_clusters_df.to_csv(os.path.join(outputs_dir, 'all_clusters.csv'), index=False)
    else:
        cluster_sorted_indices = None

    # render boxplots showing DNN scores across all clusters
    #fig, ax = plt.subplots(figsize=(10,5))
    if len(boxplot_data) <= 100:
        box = plt.boxplot(boxplot_data[::-1], vert=False, showfliers=False, medianprops={'color': 'black'})
        plt.yticks([i for i in range(1, len(boxplot_data)+1)], [str(i) for i in range(len(boxplot_data))][::-1], fontsize=4)
    else:
        for pos, values in enumerate(boxplot_data):
            values = np.array(values)            
            median = np.median(values)
            q1 = np.percentile(values, 25)
            q3 = np.percentile(values, 75)
            plt.plot([q1, q3], [pos, pos], color='gray', lw=.5) # plot the IQR line
            plt.plot(median, pos, 'o', color='k', markersize=1, zorder=100) # plot the median point
        plt.yticks([])
        plt.gca().invert_yaxis()
    plt.ylabel('Clusters')
    plt.xlabel('DNN')
    plt.gca().xaxis.set_major_locator(MaxNLocator(integer=True))
    plt.title('Average IQR: %.2f' % average_iqr)
    if ref_seq is not None:
        if sort_median is True:
            ref_cluster = cluster_sorted_indices.index(all_clusters_df.loc[ref_idx, 'Cluster'])
        elif sort_median is False and subset_clusters is False:
            ref_cluster = all_clusters_df.loc[ref_idx, 'Cluster']
        plt.axvline(np.median(boxplot_data[ref_cluster]), c='red', label='Ref', zorder=-100)
        plt.legend(loc='best')
    plt.tight_layout()
    plt.savefig(os.path.join(outputs_dir, 'clusters_boxplot.png' % k), facecolor='w', dpi=dpi)
    plt.cla() 
    plt.clf()
    plt.close()

    if 1: # render plots used in manuscript figures
        # plot just the median values for clusters in a bar plot
        bar_height = 0.8#1.0 #0.6
        median_values = []
        for i in range(len(boxplot_data)):
            median_values.append(np.median(boxplot_data[i]))
        plt.figure(figsize=(1,5))
        y_positions = np.arange(len(median_values))
        if ref_seq is not None:
            for i in range(len(median_values)):
                if i == ref_cluster:
                    plt.barh(i, median_values[i], color='red', height=bar_height)
                else:
                    plt.barh(i, median_values[i], color='C0', height=bar_height)
        else:
            plt.barh(y_positions, median_values, height=bar_height)
        plt.gca().invert_yaxis()
        plt.axvline(x=0, color='black', linewidth=.5, zorder=100)
        #plt.xlim(np.amin(median_values),np.amax(median_values))

        plt.yticks(y_positions, y_positions)

        plt.tight_layout()
        plt.savefig(os.path.join(outputs_dir, 'clusters_median_barplot.png'), facecolor='w', dpi=dpi)
        plt.cla()
        plt.clf()
        plt.close()

        # plot the occupancy of each cluster in a bar plot
        occupancy_values = []
        for i in range(len(boxplot_data)):
            occupancy_values.append(len(boxplot_data[i]))
        occupancy_values = occupancy_values[::-1]
        plt.figure(figsize=(1.5,5))
        if ref_seq is not None:
            for i in range(len(occupancy_values)):
                if i == ref_cluster:
                    plt.barh(i, occupancy_values[i], color='red', height=1.0)
                else:
                    plt.barh(i, occupancy_values[i], color='C0', height=1.0)
        else:
            plt.barh(y_positions, occupancy_values, color='C0', height=1.0)

        plt.gca().invert_yaxis()
        plt.axvline(x=0, color='black', linewidth=.5, zorder=100)
        plt.tight_layout()
        plt.savefig(os.path.join(outputs_dir, 'clusters_occupancy_barplot.png'), facecolor='w', dpi=dpi)
        plt.cla()
        plt.clf()
        plt.close()


if 1:
    print('Computing sequence statistics...')
    regulatory_df = generate_regulatory_df(mave, clusters_labels, clusters_idx, ref_seq)

    regulatory_df.to_csv(os.path.join(outputs_dir, 'compare_clusters.csv'), index=False)

    if 1:
        regulatory_df_csv = pd.read_csv(os.path.join(outputs_dir, 'compare_clusters.csv'))
        # plot regulatory matrix using different metrics:
        fig = plt.figure(figsize=(20, 5))
        ax, cax, reordered_ind, revels = impress.plot_clusters_matches_2d_gui(regulatory_df_csv, fig, sort=sort_fig, column='Entropy', sort_index=cluster_sorted_indices)
        ax.set_title('')
        plt.savefig(os.path.join(outputs_dir, 'compare_clusters_entropy.png'), facecolor='w', dpi=dpi)
        plt.close()
        if ref_seq is not None:
            fig = plt.figure(figsize=(20, 5))
            ax, cax, reordered_ind, revels = impress.plot_clusters_matches_2d_gui(regulatory_df_csv, fig, sort=sort_fig, column='Reference', sort_index=cluster_sorted_indices)
            ax.set_title('')
            plt.savefig(os.path.join(outputs_dir, 'compare_clusters_reference.png'), facecolor='w', dpi=dpi)
            plt.close()


if 1: # create percent mismatch table
    if ref_seq is not None:
        print('Calculating percent mismatch table...')
        mismatch_df = impress.mismatch2csv(regulatory_df_csv, all_clusters_df, mave, ref_seq,
                                        threshold=90, # minimum percent mismatch at any position
                                        sort_index=cluster_sorted_indices,
                                        alphabet=alphabet, save_dir=outputs_dir)


if 1: # render logos for all clusters

    if seq_logo_pwm is True or seq_logo_enrich is True:
        if not os.path.exists(os.path.join(outputs_dir, 'clusters_seq')):
            os.mkdir(os.path.join(outputs_dir, 'clusters_seq'))
    if bias_removal is False:
        clusters_avg_fname = 'clusters_avg'
        if not os.path.exists(os.path.join(outputs_dir, clusters_avg_fname)):
            os.mkdir(os.path.join(outputs_dir, clusters_avg_fname))
    elif bias_removal is True:
        clusters_bias_fname = 'clusters_bias'
        if not os.path.exists(os.path.join(outputs_dir, clusters_bias_fname)):
            os.mkdir(os.path.join(outputs_dir, clusters_bias_fname))
        clusters_avg_fname = 'clusters_avg_nobias'
        if not os.path.exists(os.path.join(outputs_dir, clusters_avg_fname)):
            os.mkdir(os.path.join(outputs_dir, clusters_avg_fname))

    if plot_profiles is True:
        y_profiles = np.load(os.path.join(inputs_dir, 'y_profiles.npy'))
        if not os.path.exists(os.path.join(outputs_dir, 'clusters_profiles')):
            os.mkdir(os.path.join(outputs_dir, 'clusters_profiles'))

    if attr_logo_fixed is True:
        print('Finding global y_min/max of attribution maps...')
        y_min, y_max = 0., 0.
        for k in tqdm(clusters_idx, desc='Clusters'):
            k_idxs =  mave.loc[mave['Cluster'] == k].index
            map_avg = np.mean(maps[k_idxs], axis=0)
            map_avg -= np.expand_dims(np.mean(map_avg, axis=1), axis=1) # position-dependent centering (gauge fix)

            positive_mask = map_avg > 0
            positive_matrix = map_avg * positive_mask
            positive_summations = positive_matrix.sum(axis=1)
            positive_max_position = np.argmax(positive_summations)
            if positive_summations[positive_max_position] > y_max:
                y_max = positive_summations[positive_max_position]

            negative_mask = map_avg < 0
            negative_matrix = map_avg * negative_mask
            negative_summations = negative_matrix.sum(axis=1)
            negative_max_position = np.argmin(negative_summations)
            if negative_summations[negative_max_position] < y_min:
                y_min = negative_summations[negative_max_position]

    if bias_removal is True:
        null_rate = 1 - mut_rate
        background_entropy = entropy(np.array([null_rate, (1-null_rate)/3, (1-null_rate)/3,(1-null_rate)/3]), base=2) #e.g., 0.096 for r=0.01
        entropy_threshold = background_entropy*0.5
        cluster_bias = np.zeros(shape=(len(clusters_idx), maps.shape[1], maps.shape[2]))
        k_name = 0 # necessary for keeping logo filenames in correct order if using 'sort_median' above
        for k in tqdm(clusters_idx, desc='Bias'):
            k_idxs = mave.loc[mave['Cluster'] == k].index
            child_maps = maps[k_idxs]
            entropic_positions = regulatory_df[(regulatory_df['Cluster'] == k) & 
                                            (regulatory_df['Entropy'] > entropy_threshold)]['Position'].values
            if len(entropic_positions) == 0:
                continue
            for ep in entropic_positions:
                for child in child_maps:
                    cluster_bias[k,ep,:] += child[ep,:]
            cluster_bias[k] /= len(child_maps)

            if plot_bias is True:
                try:
                    cluster_bias_map = squid_utils.arr2pd(cluster_bias[k], alphabet)
                except:
                    cluster_bias_map = squid_utils.arr2pd(cluster_bias[k]) # ZULU: may need a separate option for this to deal with two-hot encodings (e.g., CLIPNET)
                if attr_logo_fixed is True:
                    y_min_max = [y_min, y_max]
                else:
                    y_min_max = None
                logo_fig = impress.plot_logo(cluster_bias_map, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl, figsize=figsize, y_min_max=y_min_max)
                plt.savefig(os.path.join(outputs_dir, 'clusters_bias/cluster_%s.png' % k_name), facecolor='w', dpi=dpi)
                plt.close()
            k_name += 1
        all_bias = np.mean(cluster_bias, axis=0)
        np.save(os.path.join(outputs_dir, 'bias_matrix.npy'), all_bias)


    if attr_logo_fixed is True: # render individual reference map using global y_min/max
        ref_map = np.copy(maps[ref_idx])
        if bias_removal is True:
            ref_map -= all_bias
        try:
            ref_map = squid_utils.arr2pd(ref_map, alphabet)
        except:
            ref_map = squid_utils.arr2pd(ref_map) # ZULU: may need a separate option for this to deal with two-hot encodings (e.g., CLIPNET)
        logo_fig = impress.plot_logo(ref_map, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl, y_min_max=[y_min, y_max], figsize=figsize)
        if bias_removal is False and ref_seq is not None:
            plt.savefig(os.path.join(outputs_dir, 'individual_logo_reference_y_fixed.png'), facecolor='w', dpi=dpi)
        elif bias_removal is True and ref_seq is not None:
            plt.savefig(os.path.join(outputs_dir, 'individual_logo_reference_y_fixed_nobias.png'), facecolor='w', dpi=dpi)
        plt.close()
        if 1: # render the map with the highest prediction score using global y_min/max (useful for optimization analysis)
            max_idx = mave['DNN'].argmax()
            print('Max idx:', max_idx)
            max_map = squid_utils.arr2pd(maps[max_idx], alphabet)
            logo_fig = impress.plot_logo(max_map, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl, y_min_max=[y_min, y_max], figsize=figsize)
            plt.savefig(os.path.join(outputs_dir, 'individual_logo_maxpred.png'), facecolor='w', dpi=dpi)
            plt.close()
        if 1: # render the average of all maps
            try:
                all_maps_avg = squid_utils.arr2pd(np.mean(maps, axis=0), alphabet)
            except: #ZULU
                all_maps_avg = squid_utils.arr2pd(np.mean(maps, axis=0))
            logo_fig = impress.plot_logo(all_maps_avg, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl, y_min_max=[y_min, y_max], figsize=figsize)
            plt.savefig(os.path.join(outputs_dir, 'all_maps_average_y_fixed.png'), facecolor='w', dpi=dpi)
            plt.close()


    if attr_logo_adaptive is True: # render individual reference map using adaptive y-axis
        ref_map = np.copy(maps[ref_idx])
        if bias_removal is True:
            ref_map -= all_bias
        try:
            ref_map = squid_utils.arr2pd(ref_map, alphabet)
        except:
            ref_map = squid_utils.arr2pd(ref_map) # ZULU: may need a separate option for this to deal with two-hot encodings (e.g., CLIPNET)
        logo_fig = impress.plot_logo(ref_map, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl, figsize=figsize)
        if bias_removal is False and ref_seq is not None:
            plt.savefig(os.path.join(outputs_dir, 'individual_logo_reference_y_adaptive.png'), facecolor='w', dpi=dpi)
        elif bias_removal is True and ref_seq is not None:
            plt.savefig(os.path.join(outputs_dir, 'individual_logo_reference_y_adaptive_nobias.png'), facecolor='w', dpi=dpi)
        plt.close()

        if 1: # render the average of all maps
            try:
                all_maps_avg = squid_utils.arr2pd(np.mean(maps, axis=0), alphabet)
            except: #ZULU
                all_maps_avg = squid_utils.arr2pd(np.mean(maps, axis=0))
            logo_fig = impress.plot_logo(all_maps_avg, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl, figsize=figsize)
            plt.savefig(os.path.join(outputs_dir, 'all_maps_average_y_adaptive.png'), facecolor='w', dpi=dpi)
            plt.close()

    if bias_removal is True:
        try:
            bias_map = squid_utils.arr2pd(all_bias, alphabet)
        except:
            bias_map = squid_utils.arr2pd(all_bias) # ZULU: may need a separate option for this to deal with two-hot encodings (e.g., CLIPNET)
        if attr_logo_adaptive is True:
            logo_fig = impress.plot_logo(bias_map, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl, figsize=figsize)
            plt.savefig(os.path.join(outputs_dir, 'bias_y_adaptive.png'), facecolor='w', dpi=dpi)
            plt.close()
        if attr_logo_fixed is True:
            logo_fig = impress.plot_logo(bias_map, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl,  y_min_max=[y_min, y_max], figsize=figsize)
            plt.savefig(os.path.join(outputs_dir, 'bias_y_fixed.png'), facecolor='w', dpi=dpi)
            plt.close()

    print('Rendering logos...')

    consensus_df = pd.DataFrame(columns = ['Cluster','Median','Mean','Centroid','Consensus'], index=range(len(clusters_idx))) # records useful information for later review
    clusters_avg_matrices = np.zeros(shape=(len(clusters_idx), maps.shape[1], maps.shape[2]))
    k_name = 0 # necessary for keeping logo filenames in correct order if using 'sort_median' above
    for k in tqdm(clusters_idx, desc='Clusters'):
        if k >= 0: # this line is a placeholder for optionally inserting code for a specific cluster
            k_idxs = mave.loc[mave['Cluster'] == k].index
            seqs_k = mave['Sequence'][k_idxs]

            # save consensus sequence of each cluster to dataframe
            seq_array_cluster = motifs.create(seqs_k, alphabet=alphabet)
            pfm_cluster = seq_array_cluster.counts # position frequency matrix
            consensus_df.at[k_name, 'Cluster'] = k_name
            consensus_seq = pfm_cluster.consensus
            consensus_df.at[k_name, 'Consensus'] = consensus_seq
            # save the centroid sequence of each cluster to dataframe
            hamming_distances = [utilities.hamming_distance(consensus_seq, seq) for seq in seqs_k]
            centroid_index = np.argmin(hamming_distances)
            centroid_seq = seqs_k.iloc[centroid_index]
            consensus_df.at[k_name, 'Centroid'] = centroid_seq
            try:
                consensus_df.at[k_name, 'Median'] = np.median(boxplot_data[k_name])
                consensus_df.at[k_name, 'Mean'] = np.mean(boxplot_data[k_name])
            except:
                pass
        
            if plot_profiles is True:
                fig, ax = plt.subplots(figsize=(20,1.2))#.5))
                for i in k_idxs:
                    ax.plot(y_profiles[i,0],
                        color='C0', linewidth=.5, zorder=0, alpha=0.1)
                    ax.plot(-1.*y_profiles[i,1],
                            color='C3', linewidth=.5, zorder=-10, alpha=0.1)
                #if profiles_ylim is not None:
                #    ax.set_ylim(profiles_ylim)
                plt.tight_layout()
                plt.xlim(250, 750)
                plt.savefig(os.path.join(outputs_dir, 'clusters_profiles/profiles_%s.png' % k_name), dpi=200)
                plt.close()

            if seq_logo_pwm is True or seq_logo_enrich is True:
                seq_array_cluster = motifs.create(seqs_k, alphabet=alphabet)
                pfm_cluster = seq_array_cluster.counts # position frequency matrix
                pseudocounts = 0.5
                pwm_cluster = pfm_cluster.normalize(pseudocounts=pseudocounts) #https://biopython-tutorial.readthedocs.io/en/latest/notebooks/14%20-%20Sequence%20motif%20analysis%20using%20Bio.motifs.html
                if seq_logo_pwm is True:
                    if 0: # standard PWM
                        seq_logo = pd.DataFrame(pwm_cluster)
                    else: # compute Position Probability Matrix (PPM) and compute information content
                        ppm_df = pd.DataFrame(pwm_cluster)#, index=list(alphabet))
                        background = np.array([1.0 / len(alphabet)] * len(alphabet))  # assuming uniform background
                        ppm_df += 1e-6  # to avoid log(0), add a small value
                        info_content = np.sum(ppm_df * np.log2(ppm_df / background), axis=1)# + np.log2(4)
                        seq_logo = ppm_df.multiply(info_content, axis=0)

                if 1:
                    logo_fig, axis = impress.plot_logo(seq_logo, logo_type='sequence', ref_seq=None, aesthetic_lvl=aesthetic_lvl, center_values=False, figsize=figsize)
                    logo_dir = os.path.join(outputs_dir, 'clusters_seq/pwm')
                    if not os.path.exists(logo_dir):
                        os.mkdir(logo_dir)
                    plt.savefig(os.path.join(logo_dir, 'cluster_%s.png' % k_name), facecolor='w', dpi=dpi)
                    plt.cla()
                    plt.clf()
                    plt.close()

                if seq_logo_enrich is True: # enrichment logo (see https://logomaker.readthedocs.io/en/latest/examples.html#ars-enrichment-logo)
                    enrichment_ratio = (pd.DataFrame(pfm_cluster) + pseudocounts) / (pd.DataFrame(pfm_background) + pseudocounts)
                    seq_logo = np.log2(enrichment_ratio)
                    logo_fig, axis = impress.plot_logo(seq_logo, logo_type='sequence', ref_seq=ref_seq, aesthetic_lvl=aesthetic_lvl, figsize=figsize)
                    logo_dir = os.path.join(outputs_dir, 'clusters_seq/enrich')
                    if not os.path.exists(logo_dir):
                        os.mkdir(logo_dir)
                    plt.savefig(os.path.join(logo_dir, 'cluster_%s.png' % k_name), facecolor='w', dpi=dpi)
                    plt.cla() 
                    plt.clf()
                    plt.close()

            if attr_logo_fixed is True or attr_logo_adaptive is True:
                maps_avg = np.mean(maps[k_idxs], axis=0)
                if bias_removal is True:
                    maps_avg -= all_bias

                clusters_avg_matrices[k_name] = maps_avg

                try:
                    maps_avg = squid_utils.arr2pd(maps_avg, alphabet)
                except:
                    maps_avg = squid_utils.arr2pd(maps_avg) # ZULU: may need a separate option for this to deal with two-hot encodings (e.g., CLIPNET)                
                if attr_logo_adaptive is True:
                    logo_fig, ax = impress.plot_logo(maps_avg, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl, figsize=figsize)
                    logo_dir = os.path.join(outputs_dir, '%s/y_adaptive' % clusters_avg_fname)
                    if not os.path.exists(logo_dir):
                        os.mkdir(logo_dir)
                    plt.savefig(os.path.join(logo_dir, 'cluster_%s.png' % k_name), facecolor='w', dpi=dpi)
                    plt.cla()
                    plt.clf()
                    plt.close()
                if attr_logo_fixed is True:
                    if 1:
                        logo_fig, axis = impress.plot_logo(maps_avg, logo_type='attribution', ref_seq=None, aesthetic_lvl=aesthetic_lvl, y_min_max=[y_min, y_max], figsize=figsize)
                    else: # color attribution logo using wild-type sequence
                        logo_fig, axis = impress.plot_logo(maps_avg, logo_type='sequence', ref_seq=ref_seq, aesthetic_lvl=aesthetic_lvl, y_min_max=[y_min, y_max], figsize=figsize)
                    logo_dir = os.path.join(outputs_dir, '%s/y_fixed' % clusters_avg_fname)
                    if not os.path.exists(logo_dir):
                        os.mkdir(logo_dir)
                    plt.savefig(os.path.join(logo_dir, 'cluster_%s.png' % k_name), facecolor='w', dpi=dpi)
                    plt.cla() 
                    plt.clf()
                    plt.close()

        k_name += 1
        if k%10==0:
            plt.close('all')
            gc.collect()
    plt.close()
    consensus_df.to_csv(os.path.join(outputs_dir, 'clusters_consensus.csv'), index=False)
    np.save(os.path.join(outputs_dir, '%s/clusters_avg_matrices.npy' % clusters_avg_fname), clusters_avg_matrices)

    if 1:
        if attr_logo_fixed is True:
            print('Rendering variability logo...')
            from PIL import Image

            def darken_blend(image1, image2):
                '''
                Blend two images using the 'Darken' blend mode. The 'Darken' blend mode
                compares each pixel of two images and takes the darker value of each pixel.
                '''
                arr1 = np.array(image1)
                arr2 = np.array(image2)
                darkened_array = np.minimum(arr1, arr2)
                return Image.fromarray(darkened_array)

            # overlay all PNG images in the specified folder using 'Darken' mode.
            images = []
            logo_dir = os.path.join(outputs_dir, '%s/y_fixed' % clusters_avg_fname)
            for file_name in os.listdir(logo_dir): # load all PNG images from the folder
                if file_name.endswith('.png'):
                    image_path = os.path.join(logo_dir, file_name)
                    images.append(Image.open(image_path))

            if images:
                result_image = images[0] # initialize the result with the first image
                for image in tqdm(images[1:], desc='Image'): # apply the darken blend mode to all images
                    result_image = darken_blend(result_image, image)

                if bias_removal is False:
                    result_image.save(os.path.join(outputs_dir, 'variability_logo.png'))
                elif bias_removal is True:
                    result_image.save(os.path.join(outputs_dir, 'variability_logo_nobias.png'))
                print('Variability logo saved.')
            else:
                print('No PNG images found in the specified folder.')